<?php

namespace Drupal\entity_translate_side_by_side\Form;

use Drupal\Core\Entity\Display\EntityFormDisplayInterface;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form for side-by-side translation of entities.
 */
class EntityTranslateSideBySideForm extends FormBase {

  /**
   * Fields that should not be displayed or processed.
   *
   * TODO: Consider making this blacklist configurable for the contribution module.
   *  This would allow administrators to customize which fields should be excluded
   *  from processing in the Entity Translate Side by Side module.
   *  Issue: https://www.drupal.org/project/entity_translate_side_by_side/issues/3416987
   *
   * @var string[]
   */
  protected array $blacklist = [
    'langcode',
    'uid',
    'status',
    'created',
    'changed',
    'default_langcode',
    'content_translation_source',
    'content_translation_outdated',
  ];

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entity_type_manager;

  /**
   * The language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected LanguageManagerInterface $language_manager;

  /**
   * The system default language.
   *
   * @var string
   */
  protected string $default_language;

  /**
   * Constructs a new EntityTranslateSideBySideForm object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, LanguageManagerInterface $language_manager) {
    $this->entity_type_manager = $entity_type_manager;
    $this->language_manager = $language_manager;
    $this->default_language = $this->language_manager->getDefaultLanguage()->getId();
  }

  /**
   * {@inheritdoc}
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   The service container.
   *
   * @return EntityTranslateSideBySideForm|static
   *   A new instance of the EntityTranslateSideBySideForm.
   */
  public static function create(ContainerInterface $container): EntityTranslateSideBySideForm|static {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('language_manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'entity_translate_side_by_side_form';
  }

  /**
   * Builds the translation form with side-by-side layout for different languages.
   *
   * @param array<string, mixed> $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param string|null $entity_type
   *   The entity type being translated.
   * @param int|null $entity_id
   *   The ID of the entity being translated.
   *
   * @return array<string, mixed>
   *   The form structure.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function buildForm(array $form, FormStateInterface $form_state, string $entity_type = NULL, int $entity_id = NULL): array {
    // Check whether the entity type exists
    if (!$this->entity_type_manager->hasDefinition($entity_type)) {
      $this->messenger()->addError($this->t('The entity type @entity_type does not exist.', ['@entity_type' => $entity_type]));
      return $form;
    }

    // Load the entity based on type and ID.
    $entity = $this->entity_type_manager->getStorage($entity_type)
      ->load($entity_id);

    if (!$entity) {
      $this->messenger()->addError($this->t('The requested entity could not be found.'));
      return $form;
    }

    $languages = $this->language_manager->getLanguages();
    $language_options = [];
    foreach ($languages as $langcode => $language) {
      $language_options[$langcode] = $language->getName();
    }

    // Load the configuration languages or, if not available, the entity languages.
    $config = $this->config('entity_translate_side_by_side.settings');
    $default_languages = $config->get('languages');

    // If no standard languages are defined in the configuration, use the translation languages of the entity.
    if (empty($default_languages)) {
      // @phpstan-ignore-next-line
      $entity_translation_languages = array_keys($entity->getTranslationLanguages());

      // If no translation languages are available, use the default language.
      if (empty($entity_translation_languages)) {
        $default_languages = [$this->language_manager->getDefaultLanguage()->getId()];
      }
      else {
        $default_languages = $entity_translation_languages;
      }
    }

    $form['entity_type'] = [
      '#type' => 'hidden',
      '#value' => $entity_type,
    ];

    $form['entity_id'] = [
      '#type' => 'hidden',
      '#value' => $entity_id,
    ];

    $form['form_info'] = [
      '#markup' => $this->t('Select the languages you want to work with and provide translations in the fields below.'),
      '#prefix' => '<div class="info-message">',
      '#suffix' => '</div>',
    ];

    $selected_languages = \Drupal::request()->query->get('langcodes');
    if ($selected_languages) {
      $selected_languages = explode(',', $selected_languages);
    }
    else {
      $selected_languages = $default_languages;
    }

    // @phpstan-ignore-next-line
    $fields = $entity->getFieldDefinitions();

    // Create form elements for each supported language and field
    $form = $this->addFieldsForSelectedLanguages($selected_languages, $form_state, $form, $entity, $fields);
    $form = $this->getLanguageSelectDetailsSummary($form, $language_options, $selected_languages);

    if ($form['#fields_have_been_added']) {
      $form['actions']['submit'] = [
        '#type' => 'submit',
        '#value' => $this->t('Save translations'),
        '#button_type' => 'primary',
      ];
    }

    $form['#attached']['library'][] = 'entity_translate_side_by_side/edit_form_styles';
    $form['#attached']['library'][] = 'entity_translate_side_by_side/edit_form_scripts';

    return $form;
  }

  /**
   * Form submission handler.
   *
   * Processes the translation submission, saving translations for the selected languages.
   *
   * @param array<string, mixed> &$form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $entity_type = $form_state->getValue('entity_type');
    $entity_id = $form_state->getValue('entity_id');
    $entity = $this->entity_type_manager->getStorage($entity_type)->load($entity_id);

    if (!$entity) {
      $this->messenger()->addError($this->t('Unable to load the entity.'));
      return;
    }

    // @phpstan-ignore-next-line
    $default_translation = $entity->getTranslation($this->default_language);
    $is_default_translation_published = $default_translation->isPublished();

    $selected_languages = array_filter($form_state->getValue('language_selection'));
    // Track if changes were made at all.
    $changes_made = FALSE;
    $updated_languages = [];
    $updated_fields_by_language = [];

    foreach ($selected_languages as $langcode) {
      // Check whether the translation already exists.
      // @phpstan-ignore-next-line
      $is_new_translation = !$entity->hasTranslation($langcode);
      // @phpstan-ignore-next-line
      $translation = $is_new_translation ? $entity->addTranslation($langcode) : $entity->getTranslation($langcode);

      $updated_fields = [];
      // @phpstan-ignore-next-line
      foreach ($entity->getFieldDefinitions() as $field_name => $field_definition) {
        if ($this->isFieldEditable($field_definition, $field_name)) {
          // Track if changes were made at the current field level.
          $changes_at_field_level_made = FALSE;
          $new_field_value = $form_state->getValue(['language_fields', $langcode, $field_name]);
          $old_field_value = $translation->get($field_name)->getValue();

          // Check whether it is a multi-value text field.
          if ($new_field_value !== NULL && $field_definition->getType() === 'text' && $field_definition->getFieldStorageDefinition()->getCardinality() !== 1) {
            $formatted_new_values = [];

            foreach ($new_field_value as $delta => $value) {
              $existing_format = $old_field_value[$delta]['format'] ?? 'basic_html';
              $formatted_new_values[$delta] = [
                'value' => $value,
                'format' => $existing_format,
              ];
            }

            // Check whether changes have been made.
            if ($old_field_value !== $formatted_new_values) {
              $translation->set($field_name, $formatted_new_values);
              $changes_made = TRUE;
              $changes_at_field_level_made = TRUE;
            }
          }
          else {
            // It's a singe-value text field.
            // Check whether changes have been made.
            if (is_array($new_field_value) && isset($new_field_value['value'])) {
              // Associative array with 'value' and optional 'format'.
              $new_value = $new_field_value['value'];
              $new_format = $new_field_value['format'] ?? 'basic_html';
            }
            elseif (isset($new_field_value[0]['value']) && is_array($new_field_value) && is_array($new_field_value[0])) {
              // Numeric array with an associative array at index 0.
              $new_value = $new_field_value[0]['value'];
              $new_format = $new_field_value[0]['format'] ?? 'basic_html';
            }
            else {
              // Handling as a simple string.
              $new_value = $new_field_value;
              $new_format = 'basic_html';
            }

            if (is_array($old_field_value) && isset($old_field_value['value'])) {
              // Associative array with 'value' and optional 'format'.
              $old_value = $old_field_value['value'];
            }
            elseif (isset($old_field_value[0]['value']) && is_array($old_field_value) && is_array($old_field_value[0])) {
              // Numeric array with an associative array at index 0.
              $old_value = $old_field_value[0]['value'];
            }
            else {
              // Handling as a simple string.
              $old_value = $old_field_value;
            }

            if ($new_value !== $old_value) {
              $translation->set($field_name, ['value' => $new_value, 'format' => $new_format]);
              $changes_made = TRUE;
              $changes_at_field_level_made = TRUE;
            }
          }

          if ($changes_at_field_level_made) {
            $updated_fields[] = $field_name;
          }
        }
      }

      if (!empty($updated_fields)) {
        $updated_fields_by_language[$langcode] = $updated_fields;
        $updated_languages[] = $langcode;
      }

      if ($changes_made) {
        if ($is_default_translation_published) {
          // Only publish if the standard translation has been published.
          // If it is a new translation, the translation is activated by default.
          // Otherwise, the current status is retained.
          $current_status = $is_new_translation ? 1 : $translation->isPublished();
        }
        else {
          // Do not publish if the standard translation is not published.
          $current_status = 0;
        }

        $translation->set('status', $current_status);
        $translation->save();
      }
    }

    if ($changes_made) {
      $details_list = [];
      foreach ($updated_languages as $langcode) {
        $language_name = $this->language_manager->getLanguage($langcode)->getName();

        $updated_field_labels = array_map(static function ($field_name) use ($entity) {
          // @phpstan-ignore-next-line
          return $entity->getFieldDefinition($field_name)->getLabel();
        }, $updated_fields_by_language[$langcode]);

        $fields_list = implode('</li><li>', $updated_field_labels);
        $details_list[] = '<strong class="success-message-language-name-display-block">' . $language_name . ':</strong> <ul><li>' . $fields_list . '</li></ul>';
      }

      $formatted_details = Markup::create(implode($details_list));
      $this->messenger()->addMessage($this->t('Updated translations and fields: @details', ['@details' => $formatted_details]));
    }
    else {
      $this->messenger()->addMessage($this->t('No changes to save.'), 'warning');
    }
  }

  /**
   * Adds fields for selected languages to the form.
   *
   * When analyzing the entity fields, it is decided which field will be offered for translation and how it must be presented.
   * Moreover, it is determined whether translations already exist, and if they do, the stored values are displayed.
   * If no translations or values exist, empty fields are displayed according to the standard language.
   *
   * @param mixed $selected_languages
   *   The languages selected by the user for translation.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param array<string, mixed> $form
   *   The form to which fields will be added.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity being translated.
   * @param array<string, mixed> $fields
   *   The field definitions for the entity.
   *
   * @return array<string, mixed>
   *   The modified form structure.
   */
  protected function addFieldsForSelectedLanguages(mixed $selected_languages, FormStateInterface $form_state, array $form, EntityInterface $entity, array $fields): array {
    $form['language_fields'] = $this->initializeLanguageFieldsContainer();
    $sorted_fields = $this->createSortedListOfFields($entity, $fields);
    $form['#fields_have_been_added'] = FALSE;

    foreach ($selected_languages as $langcode) {
      // @phpstan-ignore-next-line
      $entity_has_translation = $entity->hasTranslation($langcode);
      $form['language_fields'][$langcode] = $this->initializeLanguageContainer($langcode, $entity_has_translation);

      // Iterate over all fields of the entity.
      foreach ($sorted_fields as $field_name => $field_definition) {
        // Check if the field is at all editable.
        if ($this->isFieldEditable($field_definition, $field_name)) {
          $form['#fields_have_been_added'] = TRUE;
          $field_type = $field_definition->getType();

          // If it is a multi-value text field, create an input for each value.
          if ($field_type === 'text' && $field_definition->getFieldStorageDefinition()->getCardinality() !== 1) {
            // @phpstan-ignore-next-line
            $values_from_default_language = $entity->getTranslation($this->default_language)->get($field_name)->getValue();
            $count_values_from_default_language = count($values_from_default_language);

            // Check whether a translation already exists.
            // If yes, create inputs for the existing translations and create empty inputs for the missing ones if required.
            // If not, empty inputs must be created for all multi-value fields.
            if ($entity_has_translation) {
              // @phpstan-ignore-next-line
              $values_from_current_language = $entity->getTranslation($langcode)->get($field_name)->getValue();
              $count_values_from_current_language = count($values_from_current_language);
              $difference = $count_values_from_default_language - $count_values_from_current_language;

              if (empty($values_from_current_language) && $count_values_from_default_language > 0) {
                // The translation exists, but is without values. We need to create all empty inputs.
                $form = $this->addInputsForNotYetTranslatedField($field_name, $field_definition, $form, $langcode, $count_values_from_default_language);
              }
              elseif (empty($values_from_current_language) && $count_values_from_default_language === 0) {
                // The translation exists, but has no values and the default translation also has no translation. We need to create all empty inputs based on the field definition.
                $form = $this->addInputsForNotYetTranslatedField($field_name, $field_definition, $form, $langcode, $field_definition->getFieldStorageDefinition()->getCardinality());
              }
              elseif ($difference > 0) {
                // Difference is positive: there were already translations, but they were not complete.
                // Create inputs for the existing fields and supplement with empty fields.
                $form = $this->addInputsForTranslatedField($values_from_current_language, $field_definition, $form, $langcode, $field_name);
                $form = $this->addInputsForNotYetTranslatedField($field_name, $field_definition, $form, $langcode, $difference, $count_values_from_current_language);
              }
              else {
                // There is no difference and all fields have values. All necessary inputs can be created with values.
                $form = $this->addInputsForTranslatedField($values_from_current_language, $field_definition, $form, $langcode, $field_name);
              }
            }
            elseif ($count_values_from_default_language > 0) {
              // There is no translation for the current language, but there is a translation for the default language. Empty entries must be created for the field.
              $form = $this->addInputsForNotYetTranslatedField($field_name, $field_definition, $form, $langcode, $count_values_from_default_language);
            }
            else {
              // There is no translation. Empty inputs must be created for all multi-value fields.
              $form = $this->addInputsForNotYetTranslatedField($field_name, $field_definition, $form, $langcode, $field_definition->getFieldStorageDefinition()->getCardinality());
            }
          }
          // Process single values.
          else {
            // Map the entity element type to the form element type.
            $form_field_type = $this->mapFieldTypeToFormType($field_type, $field_definition);
            if ($form_field_type) {
              // @phpstan-ignore-next-line
              $value = $entity_has_translation ? $entity->getTranslation($langcode)->get($field_name)->value : '';

              $form['language_fields'][$langcode][$field_name] = [
                '#type' => ($field_type === 'text_long') ? 'text_format' : $form_field_type,
                '#title' => $this->t('@field', ['@field' => $field_definition->getLabel()]),
                '#default_value' => $value,
                '#required' => $field_definition->isRequired(),
              ];

              if ($field_type === 'text_long') {
                $form['language_fields'][$langcode][$field_name]['#format'] = 'basic_html';
              }
            }
          }
        }
      }
    }

    if (!$form['#fields_have_been_added']) {
      $this->messenger()->addMessage($this->t('No fields could be loaded for translation.'), 'warning');
    }

    $form['description'] = [
      '#markup' => $this->t('You can horizontally scroll within the box to reach the desired language. Additionally, the drag-and-drop feature allows for easy navigation: simply click and hold inside the box, and then drag to the left or right to move through the different language options.'),
      '#prefix' => '<small class="scroll-box-description">',
      '#suffix' => '</small>',
    ];

    return $form;
  }

  /**
   * Builds the summary section of the form for language selection.
   *
   * @param array<string, mixed> $form
   *   The form to which the summary will be added.
   * @param array<string, mixed> $language_options
   *   The available language options.
   * @param mixed $default_languages
   *   The default selected languages.
   *
   * @return array<string, mixed>
   *   The modified form structure.
   */
  protected function getLanguageSelectDetailsSummary(array $form, array $language_options, mixed $default_languages): array {
    $form['language_details'] = [
      '#type' => 'details',
      '#title' => $this->t('Select languages to be translated'),
      '#open' => FALSE,
    ];

    $form['language_details']['language_selection'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Languages to edit'),
      '#options' => $language_options,
      '#default_value' => $default_languages,
    ];

    // This section creates a span element styled as a button specifically for updating the language selection.
    // Unlike a traditional 'submit' button, this span is designed to trigger a JavaScript function
    // rather than initiate a full form submission. This approach is advantageous in scenarios
    // where we want to refresh or rebuild the form based on certain user selections (like language preferences)
    // without going through the entire form submission process.
    //
    // Using a regular submit button would typically involve server-side processing,
    // including form validation and submission handling, which might not be necessary
    // for simply updating part of the form's content (in this case, based on selected languages).
    $form['language_details']['update'] = [
      '#markup' => '<span id="update-languages-link" class="button button--secondary">' . $this->t('Update Languages') . '</span>',
    ];

    // Add a link to the admin settings if the user has the authorization
    if (\Drupal::currentUser()
      ->hasPermission('administer site configuration')) {
      $admin_settings_url = Url::fromRoute('entity_translate_side_by_side.admin_settings');
      $form['language_details']['admin_link'] = [
        '#markup' => $this->t('Manage default language settings <a href=":url">here</a>.', [':url' => $admin_settings_url->toString()]),
        '#prefix' => '<small class="info-message-administer-site-configuration">',
        '#suffix' => '</small>',
      ];
    }
    return $form;
  }

  /**
   * Initializes the container for language fields.
   *
   * @return array<string, mixed>
   *   The container element.
   */
  protected function initializeLanguageFieldsContainer(): array {
    return [
      '#type' => 'container',
      '#attributes' => ['id' => 'languages-edit-wrapper'],
      '#tree' => TRUE,
    ];
  }

  /**
   * Initializes a container for the fields of a language.
   *
   * @param string $langcode
   *   The current language.
   * @param bool $entity_has_translation
   *   Flag whether there is already a translation for the entity.
   *
   * @return array<string, mixed>
   */
  protected function initializeLanguageContainer(string $langcode, bool $entity_has_translation): array {
    $language_name = $this->getLanguageName($langcode);
    $language_title_markup = '<h6>' . $language_name . (!$entity_has_translation ? ' (' . $this->t('New translation') . ')' : '') . '</h6>';

    return [
      '#type' => 'container',
      '#attributes' => [
        'id' => 'language-edit-wrapper-' . $langcode,
        'class' => ['language-edit-fields'],
      ],
      'language_title' => [
        '#markup' => $language_title_markup,
      ],
    ];
  }

  /**
   * Maps field types to corresponding form element types.
   *
   * TODO: Currently, in our system, mostly text fields are being translated.
   *  Other types are largely untested but noted for future implementation.
   *  These should be implemented for a final contributed version of the module.
   *  Issue: https://www.drupal.org/project/entity_translate_side_by_side/issues/3417000
   *
   * @param string $field_type
   *   The field type.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   *
   * @return string|null
   *   The form element type or null if not applicable.
   */
  protected function mapFieldTypeToFormType(string $field_type, FieldDefinitionInterface $field_definition): ?string {
    switch ($field_type) {
      case 'string':
      case 'email':
      case 'telephone':
      case 'text':
        return 'textfield';

      case 'text_long':
      case 'text_with_summary':
        return $field_definition->getSetting('text_processing') ? '' : 'textarea';

      case 'boolean':
        return 'checkbox';

      case 'list_string':
      case 'list_integer':
      case 'list_float':
        return 'select';

      case 'entity_reference':
        return 'entity_autocomplete';

      case 'datetime':
        return 'datetime';

      case 'image':
      case 'file':
        return 'managed_file';

      default:
        return NULL;
    }
  }

  /**
   * Retrieves the name of a language given its language code.
   *
   * @param string $langcode
   *   The language code.
   *
   * @return string
   *   The name of the language.
   */
  protected function getLanguageName(string $langcode): string {
    $language = $this->language_manager->getLanguage($langcode);
    return $language ? $language->getName() : 'Unknown Language';
  }

  /**
   * Determines whether a field is editable.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   * @param string $field_name
   *   The field name.
   *
   * @return bool
   *   TRUE if the field is editable, otherwise FALSE.
   */
  protected function isFieldEditable(FieldDefinitionInterface $field_definition, string $field_name): bool {
    return $field_definition->isTranslatable() &&
      !$field_definition->isReadOnly() &&
      !$field_definition->isComputed() &&
      !in_array($field_name, $this->blacklist, TRUE) &&
      !$field_definition->getFieldStorageDefinition()->isBaseField();
  }

  /**
   * Adds text fields for a not yet translated field to a form.
   *
   * This function creates a series of text fields based on the number of fields
   * to be created (`$number_of_fields_to_create`). Each text field is assigned a
   * unique title, composed of the field name and an ascending counter. The offset
   * (`$offset`) determines the starting index for the fields and the starting number for the titles.
   *
   * @param string $field_name
   *   The name of the field for which the text fields are to be created.
   * @param mixed $field_definition
   *   The definition of the field, containing details such as the label text.
   * @param array<string, mixed> $form
   *   The form array to which the fields are to be added.
   * @param mixed $langcode
   *   The language code indicating the language in which the field is to be translated.
   * @param int $number_of_fields_to_create
   *   The number of text fields to be created.
   * @param int $offset
   *   The starting index for creating the text fields. Defaults to 0.
   *   An offset of 2, for example, means that the indexing of the text fields starts at 2.
   *
   * @return array<string, mixed>
   *   The updated form array with the added text fields.
   */
  protected function addInputsForNotYetTranslatedField(string $field_name, mixed $field_definition, array $form, mixed $langcode, int $number_of_fields_to_create, int $offset = 0): array {
    for ($i = 0; $i < $number_of_fields_to_create; $i++) {
      $index = $i + $offset;
      $form['language_fields'][$langcode][$field_name][$index] = [
        '#type' => 'textfield',
        '#title' => $this->t('@field (Value @delta)', [
          '@field' => $field_definition->getLabel(),
          '@delta' => $index + 1,
        ]),
        '#default_value' => '',
        '#required' => $field_definition->isRequired(),
      ];
    }
    return $form;
  }

  /**
   * Adds text fields for a translated field to a form.
   *
   * This function iterates over the values of a field in the current language and adds corresponding
   * text fields to the form for a specified language. Each text field is populated with its respective
   * value from the current language and is labeled with the field name and a sequential number.
   *
   * @param mixed $values_from_current_language
   *   The values of the field in the current language.
   *   This can be an array of values or a single value,
   *   depending on the field type and settings.
   * @param mixed $field_definition
   *   The definition of the field.
   * @param array<string, mixed> $form
   *   The form array to which the fields are to be added.
   * @param mixed $langcode
   *   The language code indicating the language in which the field is to be translated.
   * @param string $field_name
   *   The name of the field for which the text fields are to be created.
   *
   * @return array<string, mixed>
   *   The updated form array with the added text fields.
   */
  protected function addInputsForTranslatedField(mixed $values_from_current_language, mixed $field_definition, array $form, mixed $langcode, string $field_name): array {
    foreach ($values_from_current_language as $delta => $value) {
      $form['language_fields'][$langcode][$field_name][$delta] = [
        '#type' => 'textfield',
        '#title' => $this->t('@field (Value @delta)', [
          '@field' => $field_definition->getLabel(),
          '@delta' => $delta + 1,
        ]),
        '#default_value' => $value['value'],
        '#required' => $field_definition->isRequired(),
      ];
    }
    return $form;
  }

  /**
   * Creates a sorted list of field definitions for an entity based on the form display configuration.
   *
   * This function retrieves the form display configuration for the given entity and sorts the field
   * definitions according to the order specified in the configuration (based on 'weight'). If the
   * entity's form display configuration is not available or doesn't define an order for some fields,
   * those fields are included in their original order as a fallback.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity for which the sorted list of fields is to be created.
   * @param array<string, mixed> $fields
   *   An associative array of the entity's field definitions.
   *
   * @return array<string, mixed>
   *   An associative array of the entity's field definitions sorted according to their 'weight' in
   *   the form display configuration.
   */
  protected function createSortedListOfFields(EntityInterface $entity, array $fields): array {
    if (!$entity instanceof FieldableEntityInterface) {
      return $fields;
    }

    $form_display = EntityFormDisplay::collectRenderDisplay($entity, 'default');
    if ($form_display instanceof EntityFormDisplayInterface) {
      $form_display_components = $form_display->getComponents();

      // Filter out components that are not present in the fields array
      $relevant_components = array_intersect_key($form_display_components, $fields);

      // Sort the components by weight
      uasort($relevant_components, static function ($a, $b) {
        return $a['weight'] <=> $b['weight'];
      });

      // Extract the keys (field names) from the sorted components
      $sorted_field_names = array_keys($relevant_components);

      // Create a sorted list of fields based on the sorted field names
      $sorted_fields = [];
      foreach ($sorted_field_names as $field_name) {
        if (isset($fields[$field_name])) {
          $sorted_fields[$field_name] = $fields[$field_name];
        }
      }

      return $sorted_fields;
    }

    // @phpstan-ignore-next-line
    return $fields;
  }

}
